//
//  Lynkeos
//  $Id: MyCustomViews.h 242 2005-05-17 21:37:03Z j-etienne $
//
//  Created by Jean-Etienne LAMIAUD on Mon Aug 2, 2004.
//  Copyright (c) 2003-2005. Jean-Etienne LAMIAUD
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/*!
 * @header
 * @abstract Definitions of the customized NSOutlineView
 */
#ifndef __MYCUSTOMVIEWS_H
#define __MYCUSTOMVIEWS_H

#import <Foundation/Foundation.h>

/*!
 * @class MyOutlineView
 * @abstract Outline view customized for keyDown added behaviour
 * @ingroup Views
 */
@interface MyOutlineView : NSOutlineView
@end

#endif
